;===============================================================================
	.title	Aurora 9 bar
;	
;	with hardware PWM combined with software PWM to control multiple PWM duties.
;	7 or 8 bit gamma-corrected PWM resolution
;
;	Copyright 2011 Akimitsu Sadoi All Rights Reserved
;-------------------------------------------------------------------------------
;	Version history
;	0.9	taken from Aurora 9x18
;	1.0 autoplay mode added
;===============================================================================

	.include "P24F08KA101.INC"
	.list	b=4

;===============================================================================
;	constants
;
	.global __reset
	.global __DefaultInterrupt
	.global __T2Interrupt				; Timer2 ISR entry
;-------------------------------------------------------------------------------
;
	.equ	osc_adjust,	0				; value to calibrate internal oscillator (-32 ~ +31)
	.equ	max_bor, 10					; maximum number of BOR allowed in series

	.equ	autoplay_time, 18			; autoplay time in seconds (max:134)
	.equ	autoplay_ticks, (autoplay_time*488)

	.equ	num_modes, (mode_tbl_end-mode_tbl)/2	; number of program modes

	.equ	speed_step,0x8000/8			; speed adjust step value
	.equ	speed_norm,0x7FFF			; speed adjust center/normal value

	; --- port & pin mapping for I/O -----------------------------
							;--- analog input ---
	.equ	INPUT_PORT,A	; RA0
	.equ	INPUT_PIN,1		; RA0
	.equ	INPUT_APIN,1	; analog input # - sometimes different from pin #

	; push/down
	.equ	SW_PORT, PORTA
	.equ	SW_TRIS, TRISA
	.equ	SW_PIN, 5

	; right
	.equ	SWA_PORT, PORTA
	.equ	SWA_TRIS, TRISA
	.equ	SWA_PIN, 1
	.equ	SWA_CN, 3

	; up
	.equ	SWB_PORT, PORTA
	.equ	SWB_TRIS, TRISA
	.equ	SWB_PIN, 0
	.equ	SWB_CN, 2

	; down
	.equ	SWC_PORT, PORTB
	.equ	SWC_TRIS, TRISB
	.equ	SWC_PIN, 15
	.equ	SWC_CN, 11

	; left
	.equ	SWD_PORT, PORTB
	.equ	SWD_TRIS, TRISB
	.equ	SWD_PIN, 14
	.equ	SWD_CN, 12

	.equ	debounce_time, 16			; (up to 16) x 2.048 mS
	.equ	debounce_bits, (1<<debounce_time-1)
	.equ	long_push_time, 240			; x 2.048 mS

	; LED refresh time = (pr_value+1)*(3*127+1)*Tcy = (pr_value+1)*382*Tcy
	;   ~= 4.06 mS (246 Hz) with port_delay=32, port_overhead=137
.if 1
	.equ	max_duty,	0xFF			; duty cycle value for 100% duty (8 bit)
	.equ	port_delay,	(32+10)			; delay time (Tcy) between timer INT and port set (compensates for LED drivers' fall time)
	.equ	port_overhead, (137+63)		; overhead (in Tcy) between port set and timer INT end
	.equ	port_delay_comp,(0)			; pulse rinse time compensation value
	.equ	port_blank, (24)			; blank period before start of the pulse (compensates for RGB drivers' fall time)
	.equ	pr_value,(port_delay+port_overhead)
.else	; optimized for video @225 Hz
	.equ	max_duty,	0xFF			; duty cycle value for 100% duty (8 bit)
	.equ	port_delay,	(32)			; delay time (Tcy) between timer INT and port set
	.equ	port_delay_comp,(0)			; pulse rinse time compensation value
	.equ	port_blank, (36+20)			; blank period before start of the pulse
	.equ	pr_value,(185)
.endif
	.equ	OCCON_VAL,(1<<OCM2)			; Output Compare - single pulse mode, timer2 as clock

	.equ	PORTA_bits,0
	.equ	PORTB_bits,14

	.equ	num_LEDs,	9				; number of LEDs

	; --- ports & pins for LEDs ----------------------------
						; PORT, pin#
	.equ	LED_1_PORT,B
	.equ	LED_1_PIN,2	; B2, 6

	.equ	LED_2_PORT,B
	.equ	LED_2_PIN,1	; B1, 5

	.equ	LED_3_PORT,B
	.equ	LED_3_PIN,0	; B0, 4

	.equ	LED_4_PORT,B
	.equ	LED_4_PIN,13	; B13, 16

	.equ	LED_5_PORT,B
	.equ	LED_5_PIN,12	; B12, 15

	.equ	LED_6_PORT,B
	.equ	LED_6_PIN,9	; B9, 13

	.equ	LED_7_PORT,B
	.equ	LED_7_PIN,8	; B8, 12

	.equ	LED_8_PORT,B
	.equ	LED_8_PIN,7	; B7, 11

	.equ	LED_9_PORT,B
	.equ	LED_9_PIN,4	; B4, 9

	; --- RGB ---
	.equ	LED_R_PORT,A
	.equ	LED_R_PIN,4

	.equ	LED_G_PORT,A
	.equ	LED_G_PIN,3

	.equ	LED_B_PORT,A
	.equ	LED_B_PIN,2

	.equ	PWM_PIN,6	; A6, 14

	.equ	PORTA_ON_R, 0xFFFF^(1<<LED_R_PIN)
	.equ	PORTA_ON_G, 0xFFFF^(1<<LED_G_PIN)
	.equ	PORTA_ON_B, 0xFFFF^(1<<LED_B_PIN)
	.equ	PORTA_OFF, ((1<<LED_R_PIN)|(1<<LED_G_PIN)|(1<<LED_B_PIN))	; all LEDs off

	.equ	PORTB_ON_R, 0xFFFF			; all bits on
	.equ	PORTB_ON_G, 0xFFFF			; all bits on
	.equ	PORTB_ON_B, 0xFFFF			; all bits on
	.equ	PORTB_OFF, 0				; all LEDs off

	.equ	A,0
	.equ	B,(PORTA_bits*2)

	.text
;--- LED data -> duty_buff offset lookup table ---
LED_pins:
	.byte	LED_1_PIN*2+LED_1_PORT
	.byte	LED_2_PIN*2+LED_2_PORT
	.byte	LED_3_PIN*2+LED_3_PORT
	.byte	LED_4_PIN*2+LED_4_PORT
	.byte	LED_5_PIN*2+LED_5_PORT
	.byte	LED_6_PIN*2+LED_6_PORT
	.byte	LED_7_PIN*2+LED_7_PORT
	.byte	LED_8_PIN*2+LED_8_PORT
	.byte	LED_9_PIN*2+LED_9_PORT

;===============================================================================
;	variables

	.data

bor_count:	.space 2				; BOR counter
s_flags:	.space 2				; static flags - stays put between program changes
	autoplay = 0					; autoplay mode

mode_num:	.space 2				; mode (movment) number

g_flags:	.space 2				; global flags
	do_not_read = 0					; LED data being modified - do not display
	analog = 1						; analog mode
	sample = 2						; analog mode - take sample
	HSV_mode = 3					; HSV color mode
	pause = 4						; pause animation

analog_val:	.space 2				; analog input value
analog_val_old:	.space 2			; analog input value

;--- autoplay ---
autoplay_count:	.space 2			; counter for autoplay timer

;--- button data ---
btn_data:						; button sample data
	button_flags = 0			; button flags
		btn_down = 0			; button is held down
		btn_long = 1			; long push
		btn_push = 2			; button has been pushed (long or short)
	samples = 2					; button status samples
	long_push = 4				; button long push counter

btnA_data:	.space 6
btnB_data:	.space 6
btnC_data:	.space 6
btnD_data:	.space 6
btnE_data:	.space 6


;--- animation parameters --------------
update_rate:	.space 2		; overall animation speed
update_cnt:		.space 2		; keeps track of animation update timing

anim_params:
	anim_delay = 0
	anim_duty_diff = 2
	anim_step_up = 4
	anim_step_down = 6
	anim_update_rate = 8
	anim_update = 10
	anim_release_waitH = 12	; duty_hold at high duty will be released after this wait
	anim_release_waitL = 14	; duty_hold at low duty will be released after this wait
	anim_flags = 16
		af_direction = 0	; 0=forward, 1=reverse
		af_duty_hold_rs = 1	; DUTY_HOLD_RS
	anim_release_cnt = 18	; DUTY_HOLD release counter for each LED

anim_params_R:
anim_params_H:
delay_R:		.space 2
duty_diff_R:	.space 2
step_up_R:		.space 2
step_down_R:	.space 2
update_rate_R:
update_rate_H:	.space 2
update_R:		.space 2
release_waitH_R:	.space 2
release_waitL_R:	.space 2
anim_flags_R:	.space 2
release_cnt_R:	.space 2*num_LEDs

anim_params_G:
anim_params_S:
delay_G:		.space 2
duty_diff_G:	.space 2
step_up_G:		.space 2
step_down_G:	.space 2
update_rate_G:
update_rate_S:	.space 2
update_G:		.space 2
release_waitH_G:	.space 2
release_waitL_G:	.space 2
anim_flags_G:	.space 2
release_cnt_G:	.space 2*num_LEDs

anim_params_B:
anim_params_V:
delay_B:		.space 2
duty_diff_B:	.space 2
step_up_B:		.space 2
step_down_B:	.space 2
update_rate_B:	.space 2
update_B:		.space 2
release_waitH_B:	.space 2
release_waitL_B:	.space 2
anim_flags_B:	.space 2
release_cnt_B:	.space 2*num_LEDs


;--- LED data - duty levels, and other parameters/flags ---
LED_data:
	DUTY_DIR = 13					; PWN duty sweep up/down direction (1=down)
	DUTY_HOLD = 14
LED_data_R:
LED_data_R1:	.space 2
LED_data_R2:	.space 2
LED_data_R3:	.space 2
LED_data_R4:	.space 2
LED_data_R5:	.space 2
LED_data_R6:	.space 2
LED_data_R7:	.space 2
LED_data_R8:	.space 2
LED_data_R9:	.space 2

LED_data_G:
LED_data_G1:	.space 2
LED_data_G2:	.space 2
LED_data_G3:	.space 2
LED_data_G4:	.space 2
LED_data_G5:	.space 2
LED_data_G6:	.space 2
LED_data_G7:	.space 2
LED_data_G8:	.space 2
LED_data_G9:	.space 2

LED_data_B:
LED_data_B1:	.space 2
LED_data_B2:	.space 2
LED_data_B3:	.space 2
LED_data_B4:	.space 2
LED_data_B5:	.space 2
LED_data_B6:	.space 2
LED_data_B7:	.space 2
LED_data_B8:	.space 2
LED_data_B9:	.space 2

;--- HSV data buffer ---
LED_data_H:		.space 2*num_LEDs
LED_data_S:		.space 2*num_LEDs
LED_data_V:		.space 2*num_LEDs

;--- for LED refresh/interrupt ---
RGB:			.space 2			; RGB channel counter
output_duty:	.space 2			; PWM duty level being output

duty_buff:							; LED data buffers
duty_buff_R:	.space 2*(PORTA_bits+PORTB_bits)
duty_buff_G:	.space 2*(PORTA_bits+PORTB_bits)
duty_buff_B:	.space 2*(PORTA_bits+PORTB_bits)

port_buff_R:	.space 4	; PORTA/PORTB buffers
port_buff_G:	.space 4	; PORTA/PORTB buffers
port_buff_B:	.space 4	; PORTA/PORTB buffers

;--- RGB -> HSV conversion temporary variables ---
var_I:			.space 2
var_H:			.space 2
var_1:			.space 2
var_2:			.space 2
var_3:			.space 2

end_of_vars:	.space 0						; end of variables area to initialize

;===============================================================================
;	main

	.text
__reset:
				bclr	CLKDIV,#RCDIV0			; set FRC postscaler to 0 (8 MHz)
				mov		#osc_adjust,w0			; adjust internal oscillator
				mov		w0,OSCTUN

				mov		#__SP_init,w15			; Initialize stack pointer and stack limit register,
				mov		#__SPLIM_init,w0
				mov		w0,SPLIM				;
				nop								; A NOP here is required for proper SPLIM functionality

				;--- configure interrupt ---
				bset	INTCON1,#NSTDIS			; disable nested interrupts

				mov		#psvpage(duty_LUT),w0	; set up PSV
				mov		w0,PSVPAG
				bset	CORCON,#PSV				; enable PSV

				;--- configure ports ---
				setm	AD1PCFG					; set all pins in digital mode
				mov		#PORTA_OFF,w0
				mov		w0,LATA
				mov		#PORTB_OFF,w0
				mov		w0,LATB
				clr		TRISA					; set all ports output
				clr		TRISB
				bset	SW_TRIS,#SW_PIN			; set the SW_PIN input
				bset	SWA_TRIS,#SWA_PIN		; set the SWA_PIN input
				bset	SWB_TRIS,#SWB_PIN		; set the SWB_PIN input
				bset	SWC_TRIS,#SWC_PIN		; set the SWC_PIN input
				bset	SWD_TRIS,#SWD_PIN		; set the SWD_PIN input

				bset	CNPU1,#SWA_CN			; enable the pull-ups
				bset	CNPU1,#SWB_CN
				bset	CNPU1,#SWC_CN
				bset	CNPU1,#SWD_CN

				;--- pause for apox. 100 ms ---
				clr		w0
0:				repeat	#16
				clrwdt
				dec		w0,w0
				bra		NZ,0b

				;--- check if brown-out reset --------------
				btss	RCON,#BOR
				bra		startup
bor_start:
				bclr	RCON,#BOR
				inc		bor_count				; count BOR
				mov		#max_bor,w0
				cp		bor_count
				bra		GE,power_down			; power down if too many BOR occurred
startup:
				;--- read the proir mode from EEPROM -------
				call	eeprom_read
				;--- check for autoplay --------------------
				bclr	s_flags,#autoplay		; clear autoplay mode
				cp0		w0
				bra		NN,0f					; if previous mode < 0
				bset	s_flags,#autoplay		; set autoplay mode
				clr		w0						; start from mode 0
0:
				mov		w0,mode_num
startup2:
				clrwdt

				;--- clear RAM -----------------------------
				mov		#mode_num+2,w0			; make sure that the value is even number
				disi	#(end_of_vars-(mode_num+2))*2
				repeat	#(end_of_vars-(mode_num+2))/2-1
				clr		[w0++]					; or trap will result

				mov		#speed_norm,w0				; set update_rate to the midway point
				mov		w0,update_rate

				;--- keep mode_num within range ---
				mov		#num_modes,w0			; num_modes -> w0
				cp		mode_num				; if (mode_num >= num_modes)
				bra		LTU,1f
				clr		mode_num				;   reset the mode number
1:				mov		mode_num,w0
				bra		w0
mode_tbl:
				bra		mode_01
				bra		mode_02
				bra		mode_03
				bra		mode_04
				bra		mode_05
;				bra		test_01
;				bra		test_02
;				bra		test_03
				bra		test_04
;				bra		test_05
;				bra		test_06	; analog
mode_tbl_end:

mode_01:		;---------------------------------
				; half ripple - pastel

				; set initial delay
				mov		#0,w0
				mov		w0,delay_R
				mov		#num_LEDs*13,w0
				mov		w0,delay_G
				mov		#num_LEDs*26,w0
				mov		w0,delay_B

				; set update_rate
				mov		#0xBFFF-197,w0
				mov		w0,update_rate_R
				mov		#0xBFFF,w0
				mov		w0,update_rate_G
				mov		#0xBFFF+199,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#255/num_LEDs,w0
				mov		w0,duty_diff_R
				mov		w0,duty_diff_G
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD
				bclr	LED_data_G1,#DUTY_HOLD
				bclr	LED_data_B1,#DUTY_HOLD

				bra		main

mode_02:		;---------------------------------
				; xxx colide ?

				; set initial delay
				mov		#0,w0
				mov		w0,delay_R
				mov		#num_LEDs*13,w0
				mov		w0,delay_G
				mov		#num_LEDs*26,w0
				mov		w0,delay_B

				; set update_rate
				mov		#0xBFFF-197,w0
				mov		w0,update_rate_R
				mov		#0xBFFF,w0
				mov		w0,update_rate_G
				mov		#0xBFFF+199,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#255/num_LEDs,w0
				mov		w0,duty_diff_R
				mov		w0,duty_diff_G
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD
				bclr	LED_data_G1,#DUTY_HOLD
;				bclr	LED_data_B1,#DUTY_HOLD
				bclr	LED_data_B+(2*(num_LEDs-1)),#DUTY_HOLD

				bset	anim_flags_B,#af_direction	; reverse direction

				bra		main

mode_03:		;---------------------------------
				; full ripple - more saturation

				; set update_rate
				mov		#0xFFFF*51/100,w0
				mov		w0,update_rate_R
				mov		#0xFFFF*52/100,w0
				mov		w0,update_rate_G
				mov		#0xFFFF*53/100,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#255*2/num_LEDs,w0
				mov		w0,duty_diff_R
				mov		w0,duty_diff_G
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD
				bclr	LED_data_G1+(2*num_LEDs/3),#DUTY_HOLD
				bclr	LED_data_B1+(2*2*num_LEDs/3),#DUTY_HOLD

				bra		main

mode_04:		;---------------------------------
				; almost fade

				; set initial delay
				mov		#0,w0
				mov		w0,delay_R
				mov		#567,w0
				mov		w0,delay_G
				mov		#567*2,w0
				mov		w0,delay_B

				; set update_rate
				mov		#0x5FFF-789,w0
				mov		w0,update_rate_R
				mov		#0x5FFF,w0
				mov		w0,update_rate_G
				mov		#0x5FFF+456,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#4,w0
				mov		w0,duty_diff_R
				mov		#5,w0
				mov		w0,duty_diff_G
				mov		#6,w0
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R,#DUTY_HOLD
				bclr	LED_data_G,#DUTY_HOLD
				bclr	LED_data_B,#DUTY_HOLD

				; set all DUTY_HOLD_RS bits
				bset	anim_flags_R,#af_duty_hold_rs
				bset	anim_flags_G,#af_duty_hold_rs
				bset	anim_flags_B,#af_duty_hold_rs

				; set release_wait
				mov		#1,w0
				mov		w0,release_waitH_R
				mov		w0,release_waitH_G
				mov		w0,release_waitH_B
				mov		#1,w0
				mov		w0,release_waitL_R
				mov		w0,release_waitL_G
				mov		w0,release_waitL_B

				bra		main

mode_05:		;---------------------------------
				; HSV mode test
				; R->H, G->S, B->V

				bset	g_flags,#HSV_mode			; HSV mode flag on

				; set initial delay
				mov		#0,w0
				mov		w0,delay_R
				mov		w0,delay_G
				mov		w0,delay_B

				; set update_rate
				mov		#0x2000-13,w0
				mov		w0,update_rate_R
				mov		#0xFFFF,w0
				mov		w0,update_rate_G
				mov		#0xFFFF,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		#255,w0
				mov		w0,step_up_G
				mov		#1,w0
				mov		w0,step_up_B
				; set step_down
				mov		#255,w0
				mov		w0,step_down_R
				mov		#0,w0
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#3,w0
				mov		w0,duty_diff_R
				mov		#0,w0
				mov		w0,duty_diff_G
				mov		#255/num_LEDs,w0
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data_H,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_H,#DUTY_HOLD
				bclr	LED_data_S,#DUTY_HOLD
				bclr	LED_data_V,#DUTY_HOLD

;				bset	anim_flags_G,#af_duty_hold_rs	; set DUTY_HOLD_RS bit
;				bset	anim_flags_B,#af_duty_hold_rs	; set DUTY_HOLD_RS bit

				bra		main

test_01:		;---------------------------------
				; test R

				; set initial delay
				mov		#16,w0
				mov		w0,delay_R

				; set update_rate
				mov		#0xFFFF,w0
				mov		w0,update_rate_R

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				; set step_down
				mov		w0,step_down_R

				; set duty_diff
				mov		#0xFF/num_LEDs+1,w0
				mov		w0,duty_diff_R

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD

				bset	anim_flags_R,#af_duty_hold_rs	; set DUTY_HOLD_RS bit

				bra		main

test_02:		;---------------------------------
				; test G

				; set initial delay
				mov		#16,w0
				mov		w0,delay_G

				; set update_rate
				mov		#0xFFFF,w0
				mov		w0,update_rate_G

				; set step_up
				mov		#1,w0
				mov		w0,step_up_G
				; set step_down
				mov		w0,step_down_G

				; set duty_diff
				mov		#0xFF/num_LEDs+1,w0
				mov		w0,duty_diff_G

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_G1,#DUTY_HOLD

				bset	anim_flags_G,#af_duty_hold_rs	; set DUTY_HOLD_RS bit

				bra		main

test_03:		;---------------------------------
				; test B

				; set initial delay
				mov		#16,w0
				mov		w0,delay_B

				; set update_rate
				mov		#0xFFFF,w0
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_B

				; set duty_diff
				mov		#0xFF/num_LEDs+1,w0
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_B1,#DUTY_HOLD

				; set DUTY_HOLD_RS bit
				bset	anim_flags_B,#af_duty_hold_rs

				bra		main

test_04:		;---------------------------------
				; test RGB chase

				; set initial delay
				mov		#0,w0
				mov		w0,delay_R
				mov		#256*2,w0
				mov		w0,delay_G
				mov		#256*4,w0
				mov		w0,delay_B

				; set update_rate -> max
				mov		#0xFFFF,w0
				mov		w0,update_rate_R
				mov		w0,update_rate_G
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#0xFF/num_LEDs,w0
				mov		w0,duty_diff_R
				mov		w0,duty_diff_G
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD
				bclr	LED_data_G1,#DUTY_HOLD
				bclr	LED_data_B1,#DUTY_HOLD

				; set all DUTY_HOLD_RS bits
				bset	anim_flags_R,#af_duty_hold_rs
				bset	anim_flags_G,#af_duty_hold_rs
				bset	anim_flags_B,#af_duty_hold_rs

				; set release_wait
				mov		#1,w0
				mov		w0,release_waitH_R
				mov		w0,release_waitH_G
				mov		w0,release_waitH_B
				mov		#0xFF*4,w0
				mov		w0,release_waitL_R
				mov		w0,release_waitL_G
				mov		w0,release_waitL_B

				bra		main

test_05:		;---------------------------------
				; test RGB - full spectra

				; set update_rate -> max
				mov		#0xFFFF,w0
				mov		w0,update_rate_R
				mov		w0,update_rate_G
				mov		w0,update_rate_B

				; set step_up
				mov		#1,w0
				mov		w0,step_up_R
				mov		w0,step_up_G
				mov		w0,step_up_B
				; set step_down
				mov		w0,step_down_R
				mov		w0,step_down_G
				mov		w0,step_down_B

				; set duty_diff
				mov		#(255/num_LEDs*6),w0
				mov		w0,duty_diff_R
				mov		w0,duty_diff_G
				mov		w0,duty_diff_B

				; set all DUTY_HOLD bits
				mov		#LED_data,w0
				repeat	#num_LEDs*3-1
				bset	[w0++],#DUTY_HOLD

				bclr	LED_data_R1,#DUTY_HOLD
				bclr	LED_data_G1+(2*num_LEDs/3),#DUTY_HOLD
				bclr	LED_data_B1+(2*2*num_LEDs/3),#DUTY_HOLD

				; set all DUTY_HOLD_RS bits
				bset	anim_flags_R,#af_duty_hold_rs
				bset	anim_flags_G,#af_duty_hold_rs
				bset	anim_flags_B,#af_duty_hold_rs

				; set release_wait
				mov		#(255)*2,w0
				mov		w0,release_waitH_R
				mov		w0,release_waitH_G
				mov		w0,release_waitH_B
				mov		w0,release_waitL_R
				mov		w0,release_waitL_G
				mov		w0,release_waitL_B

				bra		main

test_06:		;---------------------------------
				; analog input test

				;--- configure analog input (ADC) ----------
				bclr	AD1PCFG,#INPUT_APIN			; set input pin in analog mode
				mov		#(1<<SSRC2+1<<SSRC1+1<<SSRC0),w0
				mov		w0,AD1CON1					; auto-convert mode
				mov		#INPUT_APIN,w0				; set INPUT_APIN as CH0 input
				mov		w0,AD1CHS
				mov		#(31<<8 +2),w0				; sample time = 31 TAD, TAD = 2 Tcy
				mov		w0,AD1CON3
				bset	AD1CON1,#ADON				; turn ADC on

				bset	g_flags,#analog				; analog mode flag on

				bra		main

				;---------------------------------
				; power down
power_down:
				clr		bor_count				; clear BOR count

				bclr	T1CON,#TON				; stop timer 1
				bclr	T2CON,#TON				; stop timer 2
				bclr	IEC0,#T2IE				; disable timer2 int
				bclr	IFS0,#T2IF				; clear timer2 int flag

				mov		#PORTA_OFF,w0			; turn off all drive pins
				mov		w0,LATA
				mov		#PORTB_OFF,w0
				mov		w0,LATB

0:				clrwdt
				btss	SW_PORT,#SW_PIN			; wait until the button up
				bra		0b

0:				pwrsav	#SLEEP_MODE				; go into sleep mode
				btsc	SW_PORT,#SW_PIN			; check the button
				bra		0b						; keep sleeping until button down
0:				clrwdt
				btss	SW_PORT,#SW_PIN			; wait until the button up
				bra		0b
				bra		startup

;-------------------------------------------------------------------------------
;	main

main:
				clrwdt
				bset	RCON,#SWDTEN			; enable WDT

				;--- configure PWM -----------------------------------
				mov		#pr_value,w0
				mov		w0,PR2

				mov		#port_delay-port_delay_comp,w0
				mov		w0,OC1R					; set start time
				mov		#port_delay,w0
				mov		w0,OC1RS				; set end time (duty)
				bclr	OC1CON,#OCTSEL			; use timer 2
				bset	T2CON,#TON				; start timer 2
				bset	IEC0,#T2IE				; enable timer2 int

				;--- reset autoplay_count ----------------------------
				mov		#autoplay_ticks,w0
				mov		w0,autoplay_count

				;--- use TIMER1 as frame timebase --------------------
				mov		#0x7FFF,w0				; set PR1 = 0x7FFF
				mov		w0,PR1					; timer period = 2.048 mS
				bset	T1CON,#TON				; start timer 1

main_loop:
				btss	IFS0,#T1IF				; wait for TIMER1 overflow
				bra		main_loop
				bclr	IFS0,#T1IF				; clear T1IF

				;--- button services ---------------------------------
				call	button_check			; check the button states

				btsc	btnE_data,#btn_long		; if center button long pushed
				bra		power_down
				;--- center button ---
				btss	btnE_data,#btn_push		; if center button pushed
				bra		0f
				bclr	btnE_data,#btn_push		; clear the button flag
				btg		g_flags,#pause			; pause/un-pause animation

;				cp0		update_rate				; re-start if update_rate == 0
;				bra		Z,1f
;				clr		update_rate				; pause the animation
				bra		0f
1:												; start the animation
				mov		#speed_norm,w0			; set update_rate to the midway point
				mov		w0,update_rate
				bra		animate

0:				;--- up button ---
				btss	btnB_data,#btn_long		; if up button long pushed
				bra		0f
1:				call	button_check			; check the button states
				btsc	btnB_data,#btn_down		; wait until the button release
				bra		1b
				mov		#-1,w0					; go to autoplay mode
				bra		1f

0:				btss	btnB_data,#btn_push		; if up button pushed
				bra		0f
				bclr	btnB_data,#btn_push		; clear the button flag
				inc		mode_num				; go to next mode

				mov		mode_num,w0				; write the new mode into EEPROM
1:				call	eeprom_write
				bra		startup
0:
				;--- down button ---
				btss	btnC_data,#btn_push		; if down button pushed
				bra		0f
				bclr	btnC_data,#btn_push		; clear the button flag
				cp0		mode_num
				bra		NZ,1f
				mov		#num_modes,w0			; num_modes -> w0
				mov		w0,mode_num
1:				dec		mode_num				; go to prev mode

				mov		mode_num,w0				; write the new mode into EEPROM
				call	eeprom_write
				bra		startup
0:
				;--- right button ---
;				btsc	btnA_data,#btn_down		; if right button down
;				bra		1f
				btss	btnA_data,#btn_push		; if right button pushed
				bra		0f
				bclr	btnA_data,#btn_push		; clear the button flag
1:				; increase update_rate -> speed up
				mov		#(0xFFFF-speed_step),w0
				cp		update_rate				; skip if the speed is already max
				bra		GTU,animate
				mov		#speed_step,w0
				add		update_rate
				bra		0f
0:
				;--- left button ---
;				btsc	btnD_data,#btn_down		; if left button down
;				bra		1f
				btss	btnD_data,#btn_push		; if left button pushed
				bra		0f
				bclr	btnD_data,#btn_push		; clear the button flag
1:				; decrease update_rate -> speed down
				mov		#speed_step,w0
				cp		update_rate				; skip if the speed is already min
				bra		LEU,0f
				sub		update_rate
0:
				;--- handle pause ----
				btsc	g_flags,#pause				; if animation is paused
				bra		main_loop					; skip the rest and loop main

				;--- autoplay function -------------------------------
				btss	s_flags,#autoplay			; if autoplay mode is on
				bra		0f
				dec		autoplay_count
				bra		NZ,0f
				inc		mode_num					; go to next mode
				bra		startup2					; without EEPROM write & read
0:
animate:		;--- animate LEDs ------------------------------------
				btsc	g_flags,#analog				; if analog mode
				bra		analog_LED					; do the analog thing

				mov		update_rate,w0
				add		update_cnt
				bra		NC,main_loop				; skip update if update_cnt <= 0xFFFF

				btsc	g_flags,#HSV_mode			; if HSV mode
				bra		animate_HSV

animate_RGB:
				bset	g_flags,#do_not_read		; disable LED data read

				mov		#anim_params_R,w1			; w1 -> anim_params_R
				mov		#LED_data_R,w2				; w2 -> LED_data_R
				call	animate_LED

				mov		#anim_params_G,w1			; w1 -> anim_params_G
				mov		#LED_data_G,w2				; w2 -> LED_data_G
				call	animate_LED

				mov		#anim_params_B,w1			; w1 -> anim_params_B
				mov		#LED_data_B,w2				; w2 -> LED_data_B
				call	animate_LED

				bclr	g_flags,#do_not_read		; enable LED data read

				bra		main_loop

animate_HSV:
				mov		#anim_params_R,w1			; w1 -> anim_params_R
				mov		#LED_data_H,w2				; w2 -> LED_data_H
				call	animate_LED

				mov		#anim_params_G,w1			; w1 -> anim_params_G
				mov		#LED_data_S,w2				; w2 -> LED_data_S
				call	animate_LED

				mov		#anim_params_B,w1			; w1 -> anim_params_B
				mov		#LED_data_V,w2				; w2 -> LED_data_V
				call	animate_LED

				mov		#LED_data_H,w3				; w4 -> LED_data_H
				mov		#LED_data_R,w7				; w7 -> LED_data_R
				mov		#LED_data_R+(num_LEDs-1)*2,w9	; w9 -> end of LED_data_R

				bset	g_flags,#do_not_read		; disable LED data read

0:													; for (;w7 != w9;) {
				mov		[w3],w4
				and		#0x0FF,w4

				mov		[w3+num_LEDs*2],w5
				and		#0x0FF,w5

				mov		[w3+num_LEDs*4],w6
				and		#0x0FF,w6

				call	HSV2RGB						; do HSV -> RGB conversion
				cp		w7,w9
				bra		Z,1f
				inc2	w3,w3
				inc2	w7,w7
				bra		0b							; }
1:
				bclr	g_flags,#do_not_read		; enable LED data read
				bra		main_loop

;-------------------------------------------------------------------------------
;	HSV -> RGB conversion
;	takes:		H (0-255): w4
;				S (0-255): w5
;				V (0-255): w6
;	returns:	R (0-255): [w7] (LED_data_Rx)
;				G (0-255): [w7+num_LEDs*2] (LED_data_Gx)
;				B (0-255): [w7+num_LEDs*4] (LED_data_Bx)

HSV2RGB:
				cp0		w5							; if (S == 0) {
				bra		NZ,0f
				mov		w6,[w7]						;   R = G = B = V
				mov		w6,[w7+num_LEDs*2]
				mov		w6,[w7+num_LEDs*4]
				return
0:													; } else {
				mul.uu 	w4,#6,w0					;   var_H = H * 6
				mov		w0,var_H
				lsr		w0,#8,w0					;   var_I = var_H / 256
				mov 	w0,var_I					;   var_I (0-5)
				sl		w0,#8,w0					;   var_H = var_H - var_I*256
				sub		var_H

													;   var_1 = V(256 - S)/256
				mov		#256,w0
				sub		w0,w5,w0					;   w0 = 256 - S
				mul.uu	w0,w6,w0					;   w1:w0 = V * w0
				lsr		w0,#8,w0					;   w0 = w0 / 256
				mov		w0,var_1

													;   var_2 = V(256 - S(var_H)/256)/256
				mov		var_H,w0
				mul.uu	w0,w5,w0					;   w0 = S * var_H
				lsr		w0,#8,w0					;   w0 = w0 / 256
				neg		w0,w0						;   w0 = 256 - w0
				add		#256,w0
				mul.uu	w0,w6,w0					;   w1:w0 = V * w0
				lsr		w0,#8,w0					;   w0 = w0 / 256
				mov		w0,var_2

													;   var_3 = V(256 - S(256 - var_H)/256)/256
				mov		var_H,w0
				neg		w0,w0						;   w0 = 256 - var_H
				add		#256,w0
				mul.uu	w0,w5,w0					;   w0 = S * w0
				lsr		w0,#8,w0					;   w0 = w0 / 256
				neg		w0,w0						;   w0 = 256 - w0
				add		#256,w0
				mul.uu	w0,w6,w0					;   w1:w0 = V * w0
				lsr		w0,#8,w0					;   w0 = w0 / 256
				mov		w0,var_3

				mov		var_I,w0					;   switch (var_I)
				bra		w0
				bra		var_I0
				bra		var_I1
				bra		var_I2
				bra		var_I3
				bra		var_I4
				bra		var_I5
var_I0:
				mov		w6,w0						; R = V
				mov		w0,[w7]
				mov		var_3,w0					; G = var_3
				mov		w0,[w7+num_LEDs*2]
				mov		var_1,w0					; B = var_1
				mov		w0,[w7+num_LEDs*4]
				return
var_I1:
				mov		var_2,w0					; R = var_2
				mov		w0,[w7]
				mov		w6,w0						; G = V
				mov		w0,[w7+num_LEDs*2]
				mov		var_1,w0					; B = var_1
				mov		w0,[w7+num_LEDs*4]
				return
var_I2:
				mov		var_1,w0					; R = var_1
				mov		w0,[w7]
				mov		w6,w0						; G = V
				mov		w0,[w7+num_LEDs*2]
				mov		var_3,w0					; B = var_3
				mov		w0,[w7+num_LEDs*4]
				return
var_I3:
				mov		var_1,w0					; R = var_1
				mov		w0,[w7]
				mov		var_2,w0					; G = var_2
				mov		w0,[w7+num_LEDs*2]
				mov		w6,w0						; B = V
				mov		w0,[w7+num_LEDs*4]
				return
var_I4:
				mov		var_3,w0					; R = var_3
				mov		w0,[w7]
				mov		var_1,w0					; G = var_1
				mov		w0,[w7+num_LEDs*2]
				mov		w6,w0						; B = V
				mov		w0,[w7+num_LEDs*4]
				return
var_I5:
				mov		w6,[w7]					; R = V
				mov		var_1,w0					; G = var_1
				mov		w0,[w7+num_LEDs*2]
				mov		var_2,w0					; B = var_2
				mov		w0,[w7+num_LEDs*4]
				return
													; } (endif)

;---------------------------------------------------------------------
;	analog mode LED animation
analog_LED:
				call	analog_input
.equ	adc_floor_R,276
.equ	adc_step_R,41
.equ	adc_floor_G,279
.equ	adc_step_G,51
.equ	adc_floor_B,325
.equ	adc_step_B,81

				;--- R ---
				mov		#LED_data_R,w2				; w2 -> LED_data_R
				mov		#LED_data_R+(num_LEDs*2),w3	; w3 is loop stopper

				mov		analog_val,w0
;				lsr		analog_val,wreg
				sub		#adc_floor_R-adc_step_R,w0	; value - (adc_floor - adc_step)
1:												; while (w2 < w3) {
				mov		#max_duty,w1
				sub		#adc_step_R,w0				; value - adc_step
				bra		NN,2f						; if (w0 < 0)
				add		w1,w0,w1					;   w1 = max_duty + w0
				bra		3f							; else
2:				sub		w1,w0,w1					;   w1 = max_duty - w0
3:				bra		NN,0f						; if (w1 < 0)
				clr		w1							;   w1 = 0
0:				mov		w1,[w2++]					; result -> LED_data
				cpseq	w2,w3
				bra		1b						; } end while

				;--- G ---
				mov		#LED_data_G+(num_LEDs*2),w3	; w3 is loop stopper

				mov		analog_val,w0
;				lsr		analog_val,wreg
				sub		#adc_floor_G-adc_step_G,w0	; value - (adc_floor - adc_step)
1:												; while (w2 < w3) {
				mov		#max_duty,w1
				sub		#adc_step_G,w0				; value - adc_step
				bra		NN,2f						; if (w0 < 0)
				add		w1,w0,w1					;   w1 = max_duty + w0
				bra		3f							; else
2:				sub		w1,w0,w1					;   w1 = max_duty - w0
3:				bra		NN,0f						; if (w1 < 0)
				clr		w1							;   w1 = 0
0:				mov		w1,[w2++]					; result -> LED_data
				cpseq	w2,w3
				bra		1b						; } end while

				;--- B ---
				mov		#LED_data_B+(num_LEDs*2),w3	; w3 is loop stopper

				mov		analog_val,w0
;				lsr		analog_val,wreg
				sub		#adc_floor_B-adc_step_B,w0	; value - (adc_floor - adc_step)
1:												; while (w2 < w3) {
				mov		#max_duty,w1
				sub		#adc_step_B,w0				; value - adc_step
				bra		NN,2f						; if (w0 < 0)
				add		w1,w0,w1					;   w1 = max_duty + w0
				bra		3f							; else
2:				mov		w1,w1						;   w1 = max_duty
;2:				sub		w1,w0,w1					;   w1 = max_duty - w0
3:				bra		NN,0f						; if (w1 < 0)
				clr		w1							;   w1 = 0
0:				mov		w1,[w2++]					; result -> LED_data
				cpseq	w2,w3
				bra		1b						; } end while

				bra		main_loop

;---------------------------------------------------------------------
;	check analog input - /w slow decay
analog_input:
.if 1			; slow decay on
				btss	IFS0,#AD1IF				; return if conversion is not done
				return

				mov		ADC1BUF0,w0				; w0 = ADC value
				cp		analog_val				; if (w0 < analog_val)
				bra		LEU,0f
				dec2	analog_val				;   analog_val--
				dec2	analog_val				;   analog_val--
				bra		1f						; else
0:				add		analog_val,wreg			;   avarage it with the old value
				lsr		w0,w0
				mov		w0,analog_val
1:
				bclr	IFS0,#AD1IF				; clear the DONE flag

				return

.else			; slow decay off
				btss	IFS0,#AD1IF				; return if conversion is not done
				return

				mov		ADC1BUF0,w0				; w0 = ADC value
				add		analog_val,wreg			; avarage it with the old value
				lsr		w0,w0
				mov		w0,analog_val
1:
				bclr	IFS0,#AD1IF				; clear the DONE flag

				return
.endif
;-------------------------------------------------------------------------------
;	button check

button_check:
				;--- sample all buttons ---
				mov		SWA_PORT,w0
				btst.c	w0,#SWA_PIN
				rlc		btnA_data+samples			; add new sample at LSB

				mov		SWB_PORT,w0
				btst.c	w0,#SWB_PIN
				rlc		btnB_data+samples

				mov		SWC_PORT,w0
				btst.c	w0,#SWC_PIN
				rlc		btnC_data+samples

				mov		SWD_PORT,w0
				btst.c	w0,#SWD_PIN
				rlc		btnD_data+samples

				mov		SW_PORT,w0
				btst.c	w0,#SW_PIN
				rlc		btnE_data+samples

				mov		#5,w3						; loop # of switches (w3 as loop counter)
				mov		#btnA_data,w1
0:				call	check_one_button			; process one switch at a time
				add		w1,#(btnB_data-btnA_data),w1
				dec		w3,w3
				bra		NZ,0b

				return

;-----------------------------------------------------------
check_one_button:
				;--- process the button data ---
				btsc	[w1],#btn_down		; is the button up or down?
				bra		button_down

button_up:		;--- test if the button is down ---
				mov		#debounce_bits,w0
				mov		[w1+samples],w2
				and		w0,w2,w2					; if all samples are low
				bra		NZ,0f
				bset	[w1],#btn_down				; button is now down
				mov		#long_push_time,w0			; set the long push counter
				mov		w0,[w1+long_push]
0:				return

button_down:	;--- test if the button is up ---
				mov		#debounce_bits,w0
				mov		[w1+samples],w2
				and		w0,w2,w2					; if all samples are low
				xor		w2,w0,w2					; if all samples are high
				bra		NZ,button_held
				bclr	[w1],#btn_down				; button is now up
				bset	[w1],#btn_push				; button has been pushed
				return

button_held:
				add		w1,#long_push,w2
				dec		[w2],[w2]					; decrement long push counter
				bra		Z,0f
				return
0:				bset	[w1],#btn_long				; button is now long pushed
				mov		#long_push_time,w0			; set the long push counter
				mov		w0,[w1+long_push]
				return

;-------------------------------------------------------------------------------
;	animate LEDs
;
;	register usage: 
;		w5 = anim_release_cnt_X
;		w6 = anim_flags_X
;		w7 = max_duty - duty_diff_X
;		w8 = duty_diff_X
;		w9,w10,w11,w12,w13

animate_LED:
				add		w1,#anim_delay,w0
				cp0		[w0]						; check delay
				bra		Z,0f
				dec		[w0],[w0]
				return
0:
				mov		[w1+#anim_update_rate],w0	; update_rate_X -> w0
				add		w1,#anim_update,w3
				add		w0,[w3],[w3]				; update_X += update_rate_X
				bra		C,0f						; skip update if update_X <= 0xFFFF
				return
0:
				;--- prepare for LED loop ---
				mov		w2,w9						; w9 -> LED_data_X
				mov		#(2*(num_LEDs-1)),w0
				add		w2,w0,w11					; w11 -> last LED_data_X

				mov		[w1+#anim_step_up],w12
				mov		[w1+#anim_step_down],w13

				mov		[w1+#anim_duty_diff],w8		; w8 = duty_diff_X
				mov		#max_duty,w0
				sub		w0,w8,w7					; w7 = max_duty - duty_diff_X

				mov		[w1+#anim_flags],w6
				btsc	w6,#af_direction			; if (af_direction = 0) {
				bra		1f
				add		w2,#2,w10					;   LED_data_#2 -> w10
				bra		2f							; } else {
1:				mov		w11,w10						;   last LED_data_X -> w10
2:													; }

				add		w1,#anim_release_cnt,w5		; w5 -> release_cnt_X

LED_loop:
				;--- LED x ---
				mov		[w9],w0						; copy LED_data_Xn -> w0

				btss	w0,#DUTY_HOLD				; if DUTY_HOLD set
				bra		1f							; {
				dec		[w5],[w5]					;   if (--release_cnt_X != 0)
				bra		NZ,2f						;     continue to the next LED
				bclr	w0,#DUTY_HOLD				;   else clear DUTY_HOLD
1:													; }
				btsc	w0,#DUTY_DIR				; duty up or down?
				bra		LED_rampDown
LED_rampUp:
				cpsne.b	w8,w0						; if LED_data_Xn = duty_diff_X
				bclr	[w10],#DUTY_HOLD			; release DUTY_HOLD of next LED

				add.b	w0,w12,w0					; LED_data_Xn += step_up
				bra		NC,2f						; if LED_data_Xn >= 0xff
				setm.b	w0							;   LED_data_Xn = 0xff
				btg		w0,#DUTY_DIR				;   and toggle the dir flag
				btss	w6,#af_duty_hold_rs			;   if DUTY_HOLD_RS set
				bra		2f
				bset	w0,#DUTY_HOLD				;     hold the fade
				mov		[w1+#anim_release_waitH],w3
				mov		w3,[w5]						; reset anim_release_cnt_Xn
				bra		2f
LED_rampDown:
				cpsne.b	w7,w0						; if LED_data_Xn = (max_duty - duty_diff_X)
				bclr	[w10],#DUTY_HOLD			; release DUTY_HOLD of next LED

				sub.b	w0,w13,w0					; LED_data_Xn -= step_down
				bra		C,2f						; if LED_data_Xn <= 0
				clr.b	w0							;   LED_data_Xn = 0
				btg		w0,#DUTY_DIR				;   and toggle the dir flag
				btss	w6,#af_duty_hold_rs			;   if DUTY_HOLD_RS set
				bra		2f
				bset	w0,#DUTY_HOLD				;     hold the fade
				mov		[w1+#anim_release_waitL],w3
				mov		w3,[w5]						; reset anim_release_cnt_Xn
2:
				mov		w0,[w9]						; copy w0 -> LED_data_Xn
				cpsne	w9,w11						; last one done?
				return								;   then return

				inc2	w5,w5
				inc2	w9,w9						; else next LED
				inc2	w10,w10
				cpsgt	w10,w11						; if next LED > last LED_X
				bra		LED_loop
				mov		w2,w10						;   LED_data_#1 -> w10

				bra		LED_loop


;===============================================================================
;	PWM update service
;
__T2Interrupt:
				push.s								; push w0 - w3 and status
				push.d	w4							; push w4,w5

RGB_drive:		;--- RGB channel switch ------------------------------
				mov		RGB,w0
				bra		w0
				bra		Blanking					; first time call only
				bra		Ch_G
				bra		Ch_B
				clr		RGB		; catch

Ch_R:			;--- R channel ---------------------------------------
				;--- start single pulse generation ---
				mov		#OCCON_VAL,w0
				mov		w0,OC1CON					; set OC register
				;--- set PORTs -----------------------------
				mov		#PORTA_ON_R,w0				; load next PORTA data into W
				mov		port_buff_R,w1				; load next PORTB data into W
				mov		w0,LATA						; update PORTA LATCH
				mov		w1,LATB						; update PORTB LATCH

				;--- prepare PORT data for the next cycle --
				call	update_ports_G

				bra		RGB_done

Ch_G:			;--- G channel ---------------------------------------
				;--- start single pulse generation ---
				mov		#OCCON_VAL,w0
				mov		w0,OC1CON					; set OC register
				;--- set PORTs -----------------------------
				mov		#PORTA_ON_G,w0				; load next PORTA data into W
				mov		port_buff_G,w1				; load next PORTB data into W
				mov		w0,LATA						; update PORTA LATCH
				mov		w1,LATB						; update PORTB LATCH

				;--- prepare PORT data for the next cycle --
				call	update_ports_B

				bra		RGB_done

Ch_B:			;--- B channel ---------------------------------------
				;--- start single pulse generation ---
				mov		#OCCON_VAL,w0
				mov		w0,OC1CON					; set OC register
				;--- set PORTs -----------------------------
				mov		#PORTA_ON_B,w0				; load next PORTA data into W
				mov		port_buff_B,w1				; load next PORTB data into W
				mov		w0,LATA						; update PORTA LATCH
				mov		w1,LATB						; update PORTB LATCH

				;--- increment duty level ----------------------------
				inc2	output_duty					; inc by 2 to shorten the cycle
				mov		#max_duty,w0
				cp		output_duty					; if output_duty > max_duty
				bra		LEU,B_done
MSB_done:											; max_duty done {
				btsc	g_flags,#do_not_read		; if LED data being updated
				bra		Blanking					;   skip copying of LED data

				;--- sort & copy LED duty values to buffer -----------
				mov		#tblpage(LED_pins),w0		; prepare LED pin lookup
				mov		w0,TBLPAG
				clr		w0							; clear w0 to prepare for byte read

				;--- R data ---
				mov		#tbloffset(LED_pins),w5		; table address -> w5
				mov		#LED_data_R,w1				; w1 -> LED_data
				mov		#LED_data_R+(2*num_LEDs),w4	; w4 -> last LED_data (loop stopper)
				mov		#duty_buff_R,w2				; w2 -> duty_buff
0:
				tblrdl.b [w5++],w0					; read the LED pin destination
				mov		[w1++],[w2+w0]				; copy LED_data(w1) -> duty_buff_x
				cp		w1,w4						; finish if (w1 = w4)
				bra		NZ,0b

				;--- G data ---
				mov		#tbloffset(LED_pins),w5		; table address -> w5

				mov		#LED_data_G,w1
				mov		#LED_data_G+(2*num_LEDs),w4	; address of the last LED_data -> w4
				mov		#duty_buff_G,w2
0:
				tblrdl.b [w5++],w0					; read the LED pin destination
				mov		[w1++],[w2+w0]				; LED_data(w1) -> duty_buff_x
				cp		w1,w4						; finish if (w1 = w4)
				bra		NZ,0b

				;--- B data ---
				mov		#tbloffset(LED_pins),w5		; table address -> w5

				mov		#LED_data_B,w1
				mov		#LED_data_B+(2*num_LEDs),w4	; address of the last LED_data -> w4
				mov		#duty_buff_B,w2
0:
				tblrdl.b [w5++],w0					; read the LED pin destination
				mov		[w1++],[w2+w0]				; LED_data(w1) -> duty_buff_x
				cp		w1,w4						; finish if (w1 = w4)
				bra		NZ,0b

Blanking:		;--- Blanking period ---------------------------------
				clr		output_duty					;   output_duty = 1
				inc		output_duty
				mov		#2,w0
				mov		w0,RGB						; RGB = 2 so R-ch will be next
				;--- analog port sampling ---
				btsc	g_flags,#analog				; if analog mode on
				bset	AD1CON1,#SAMP				; start sampling analog input

				;-----------------------------------------------------
													; }
B_done:			;--- expand duty value via LUT ---
				mov		#psvoffset(duty_LUT),w3
				sl		output_duty,wreg			; address offset = value * 2
				mov		[w3+w0],w0					; w0 = pulse_duration

				;--- set next PWM duty ---
				mov		w0,OC1RS					; set the next OC1RS value

				;--- prepare PORT data for the next cycle --
				call	update_ports_R

RGB_done:		;---  ------------------------------------------------
				inc		RGB							; next RGB channel
All_done:
				clrwdt

				pop.d	w4							; pop w4,w5
				pop.s
				bclr	IFS0,#T2IF					; clear TMR2 int flag
				retfie
;
;-------------------------------------------------------------------------------
;	update ports based on the LED duty values
;
update_ports_R:
				mov		#duty_buff_R,w1			; point to duty buffer
				mov		#port_buff_R,w2			; point to port buffer
				bra		1f
update_ports_G:
				mov		#duty_buff_G,w1			; point to duty buffer
				mov		#port_buff_G,w2			; point to port buffer
				bra		1f
update_ports_B:
				mov		#duty_buff_B,w1			; point to duty buffer
				mov		#port_buff_B,w2			; point to port buffer
;				bra		1f
1:
				mov		output_duty,w3			; current output_duty -> w3
.if (PORTA_bits > 0)
				mov		#(1<<(PORTA_bits-1)),w4	; set the stopper bit
				mov		w4,[w2]					; this exits the loop
0:
				mov		[w1++],w0				; LED_data -> w0
				cp.b	w0,w3					; compare LED_data:output_duty -> C
				rrc		[w2],[w2]				; C -> MSB of port data
				bra		NC,0b
				repeat	#16-PORTA_bits-1
				lsr		[w2],[w2]
				inc2	w2
.endif
												; repeat on PORTB
				mov		#(1<<(PORTB_bits-1)),w4	; set the stopper bit
				mov		w4,[w2]					; this exits the loop
0:
				mov		[w1++],w0				; LED_data -> w0
				cp.b	w0,w3					; compare LED_data:output_duty -> C
				rrc		[w2],[w2]				; C -> MSB of port data
				bra		NC,0b
.if (PORTB_bits < 16)
				repeat	#16-PORTB_bits-1
				lsr		[w2],[w2]
.endif

				return

;-------------------------------------------------------------------------------
;	duty expansion look up table
;
	.equ	duty_knee,6
	.equ	max_pulse_length,(pr_value-port_blank-1)

duty_LUT:	.word	port_delay+1
			.rept	duty_knee			; slow slope at the low brightness
				.word	port_delay+1
			.endr
			.equ	x, 0
			.rept	max_duty-duty_knee
				.equ	x, x+1
				.equ	y, x*(max_pulse_length-1)/(max_duty-duty_knee)+1
				.word	(port_delay+y) % (pr_value+1)		; OCxRS (end of pulse) value
			.endr

.if 0	;--- show/hide pulse_duration values for evaluation ---
duty_LUT_eval:	.word	pr_value
			.rept	duty_knee			; slow slope at the low brightness
				.word	1
			.endr
			.equ	x, 0
			.rept	max_duty-duty_knee
				.equ	x, x+1
				.equ	y, x*(max_pulse_length-1)/(max_duty-duty_knee)+1
;				.equ	y, x*(max_pulse_length-1-(duty_knee/2))/(max_duty-duty_knee)+(duty_knee/2)
				.word	y
			.endr
duty_LUT_end:
			.word	max_pulse_length	; EOT mark
.endif

;-------------------------------------------------------------------------------
;	EEPROM read/write
;
	.equ	eeprom_addr,0x7ffe00	; EEPROM address to keep the data

eeprom_write:	;--- write w0 into eeprom_addr ---
				mov		#tblpage(eeprom_addr),w1	; top 8 bits of EEPROM address
				mov		w1,TBLPAG
				mov		#tbloffset(eeprom_addr),w1	; address -> w0
				tblwtl	w0,[w1]

				clr		NVMCON
				bset	NVMCON,#WREN				; set write enable bit
				bset	NVMCON,#NVMOP2

				disi	#7							; disable INITs
													; unlock sequence
				mov		#0x55,w1
				mov		w1,NVMKEY
				mov		#0xAA,w1
				mov		w1,NVMKEY
				bset	NVMCON,#WR					; initiate write oparation
				nop
				nop

0:				btsc	NVMCON,#WR					; wait till WR clears
				bra		0b

				return
				
eeprom_read:	;--- read eeprom_addr into w0
				mov		#tblpage(eeprom_addr),w0	; top 8 bits of EEPROM address
				mov		w0,TBLPAG
				mov		#tbloffset(eeprom_addr),w0	; address -> w0
				tblrdl	[w0],w0
				return

;===============================================================================
__DefaultInterrupt:
				reset

	.end
